package dk.apaq.crud.vaadin;

import com.vaadin.data.Container;
import com.vaadin.data.Item;
import com.vaadin.data.Property;
import dk.apaq.crud.Crud;
import dk.apaq.crud.CrudEvent.WithEntity;
import dk.apaq.crud.CrudEvent.WithId;
import dk.apaq.crud.CrudNotifier;
import dk.apaq.crud.HasId;
import dk.apaq.crud.core.BaseCrudListener;
import dk.apaq.filter.Filter;
import dk.apaq.filter.sort.Sorter;
import dk.apaq.filter.vaadin.FilterableContainer;
import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;

/**
 * A Vaadin Container that uses a <code>Crud</code> as a backend. The container supports the operations
 * supported by its crud while caching the object it reads in a WeakHashMap.
 *
 * - Crud.Editable -
 * If the Crud is a <code>Crud.Editable</code> instance, then it will be possible to use the addItem() and
 * removeItem() methods of the container.
 *
 * - Crud.Filterable -
 * If the Crud is an instance of <code>Crud.Filterable</code> then it will be possible to specifiy <code>Filter</code>
 * and <code>Sorter</code> on the container to have its data filtered directly in the Crud backend.
 *
 * - CrudNotifier -
 * If the Crud is an instance of <code>CrudNotifer</code> then the container will make sure to invalidate
 * changed items in the cache and reload idlist if needed.
 *
 */
public class CrudContainer<IDT, BT extends HasId<IDT>> implements
        Container, Container.Indexed, Container.Ordered, Container.ItemSetChangeNotifier,
        FilterableContainer {

    protected Sorter sorter;
    protected Filter filter;
    protected final Crud<IDT, BT> crud;
    protected List<IDT> idList = null;
    protected transient LinkedHashMap<String, Class> model;
    private final Map<IDT, CrudItem> itemIdToItem = new WeakHashMap<IDT, CrudItem>();
    private List<ItemSetChangeListener> itemSetChangeListeners = new ArrayList<ItemSetChangeListener>();
    private CrudChangeHandler crudChangeHandler = null;


    private class CrudChangeHandler extends BaseCrudListener<IDT, BT> {

        @Override
        public void onEntityUpdate(WithEntity<IDT, BT> event) {
            invalidate(event.getEntity().getId());
            updateIds();
        }

        @Override
        public void onEntityDelete(WithId<IDT, BT> event) {
            if(idList.contains(event.getEntityId())) {
                updateIds();
            }
        }

    }

    private class ChangeEvent implements ItemSetChangeEvent {

        private final Container parent;

        public ChangeEvent(Container parent) {
            this.parent = parent;
        }

        @Override
        public Container getContainer() {
            return parent;
        }
    }

    public CrudContainer(Crud<IDT, BT> crud, Class beanClass) {
        this.crud = crud;
        this.model = getPropertyDescriptors(beanClass);

        if (this.crud instanceof CrudNotifier) {
            crudChangeHandler = new CrudChangeHandler();
           ((CrudNotifier)this.crud).addListener(crudChangeHandler);
        }
    }

    protected CrudContainer(Crud<IDT, BT> crud, LinkedHashMap<String, Class> propertyMap) {
        this.crud = crud;
        this.model = propertyMap;
    }

    protected CrudItem buildItem(CrudContainer<IDT, BT> container, Crud<IDT, BT> crud, BT bean) {
        return new CrudItem(container, crud, bean);
    }

    @Override
    public Item getItem(Object itemId) {
        if(itemId==null) {
            return null;
        }
        
        CrudItem item = itemIdToItem.get((IDT)itemId);
        if(item == null) {
            item = buildItem(this, crud, crud.read((IDT)itemId));
            
            itemIdToItem.put((IDT)itemId, item);
        }
        return item;
    }

    @Override
    public Filter getFilter() {
        return filter;
    }

    @Override
    public Sorter getSorter() {
        return sorter;
    }

    @Override
    public void setFilter(Filter filter) {
        this.filter = filter;
        updateIds();
    }

    @Override
    public void setSorter(Sorter sorter) {
        this.sorter = sorter;
        updateIds();
    }

    @Override
    public void setFilterAndSorter(Filter filter, Sorter sorter) {
        this.filter = filter;
        this.sorter = sorter;
        updateIds();
    }

    /**
     * Reloads the list of ids matching the specified filter and sorter.
     */
    public void updateIds() {
        idList = getItemIdList(filter, sorter);
        fireItemSetChange();
    }

    /**
     * Loads list of ids from the crud.
     */
    protected List<IDT> getItemIdList(Filter filter, Sorter sorter) {
        if(filter != null || sorter != null) {
            if(crud instanceof Crud.Filterable) {
                    return ((Crud.Filterable)crud).listIds(filter, sorter);
            } else {
                throw new UnsupportedOperationException("Filtering not supported by the backend Crud.");
            }
        } else {
            return crud.listIds();
        }
    }

    /**
     * Invalidates a specific item.
     */
    protected void invalidate(IDT id) {
        CrudItem<IDT, BT> item = itemIdToItem.get(id);
        if(item!=null && !item.isModified()) {
            item.setInnerBean(crud.read(id));
        }
    }

    @Override
    public void addListener(ItemSetChangeListener listener) {
        itemSetChangeListeners.add(listener);
    }

    @Override
    public void removeListener(ItemSetChangeListener listener) {
        itemSetChangeListeners.remove(listener);
    }

    /**
     * Tells listeners that set of items in the container has changed.
     */
    private void fireItemSetChange() {
        ChangeEvent event = new ChangeEvent(this);
        for(ItemSetChangeListener listener : itemSetChangeListeners) {
            listener.containerItemSetChange(event);
        }
    }
    

    @Override
    public boolean addContainerProperty(Object propertyId, Class<?> type, Object defaultValue) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported by CrudContainers");
    }

    @Override
    public Object addItem() throws UnsupportedOperationException {
        if(crud instanceof Crud.Editable) {
            HasId entity = ((Crud.Editable)crud).create();
            updateIds();
            return entity.getId();
        } else {
            throw new UnsupportedOperationException("Editing not supported by the backend Crud.");
        }
    }

    @Override
    public Item addItem(Object itemId) throws UnsupportedOperationException {
        return null;
    }

    
    @Override
    public Object addItemAfter(Object previousItemId) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Item addItemAfter(Object previousItemId, Object newItemId) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Object addItemAt(int index) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Item addItemAt(int index, Object newItemId) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public boolean containsId(Object itemId) {
        return getItemIdsAsList().contains(itemId);
    }

    @Override
    public Object firstItemId() {
        List idlist = getItemIdsAsList();
        if (idlist.isEmpty()) {
            return null;
        } else {
            return idlist.get(0);
        }
    }

    @Override
    public Object getIdByIndex(int index) {
        return getItemIdsAsList().get(index);
    }

    @Override
    public Collection<?> getItemIds() {
        return Collections.unmodifiableCollection(getItemIdsAsList());
    }

    public List<?> getItemIdsAsList() {
        if (idList == null) {
            updateIds();
        }
        return idList;
    }

    @Override
    public Collection<?> getContainerPropertyIds() {
        return model.keySet();
    }

    @Override
    public Property getContainerProperty(Object itemId, Object propertyId) {
        Item item = getItem(itemId);
        return item.getItemProperty(propertyId);
    }

    @Override
    public Class<?> getType(Object propertyId) {
        Class c = model.get((String)propertyId);
        if(c == null) {
            return Object.class;
        } else {
            return c;
        }
    }

    @Override
    public int indexOfId(Object itemId) {
        return getItemIdsAsList().indexOf(itemId);
    }

    @Override
    public boolean isFirstId(Object itemId) {
        List idlist = getItemIdsAsList();
        if (idlist.isEmpty()) {
            return false;
        } else {
            return itemId.equals(idlist.get(0));
        }
    }

    @Override
    public boolean isLastId(Object itemId) {
        List idlist = getItemIdsAsList();
        if (idlist.isEmpty()) {
            return false;
        } else {
            return itemId.equals(idlist.get(idlist.size() - 1));
        }
    }

    @Override
    public Object lastItemId() {
        List idlist = getItemIdsAsList();
        if (idlist.isEmpty()) {
            return null;
        } else {
            return idlist.get(idlist.size() - 1);
        }
    }

    @Override
    public Object nextItemId(Object itemId) {
        List idlist = getItemIdsAsList();
        int index = idlist.indexOf(itemId);
        if (index >= 0 && index < idList.size() - 1) {
            return idlist.get(index + 1);
        }
        return null;
    }

    @Override
    public Object prevItemId(Object itemId) {
        List idlist = getItemIdsAsList();
        int index = idlist.indexOf(itemId);
        if (index > 0) {
            return idlist.get(index - 1);
        }
        return null;
    }

    @Override
    public int size() {
        return getItemIdsAsList().size();
    }

    @Override
    public boolean removeItem(Object itemId) throws UnsupportedOperationException {
        if(crud instanceof Crud.Editable) {
            Crud.Editable editableCrud = (Crud.Editable)crud;
            editableCrud.delete(itemId);
            updateIds();
            return true;
        } else {
            throw new UnsupportedOperationException("Editing not supported by Crud.");
        }
    }

    @Override
    public boolean removeContainerProperty(Object propertyId) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public boolean removeAllItems() throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    /**
     * <p>
     * Perform introspection on a Java Bean class to find its properties.
     * </p>
     *
     * <p>
     * Note : This version only supports introspectable bean properties and
     * their getter and setter methods. Stand-alone <code>is</code> and
     * <code>are</code> methods are not supported.
     * </p>
     *
     * @param beanClass
     *            the Java Bean class to get properties for.
     * @return an ordered map from property names to property descriptors
     */
    public static LinkedHashMap<String, Class> getPropertyDescriptors(
            final Class<?> beanClass) {
        final LinkedHashMap<String, Class> pdMap = new LinkedHashMap<String, Class>();

        // Try to introspect, if it fails, we just have an empty Item
        try {
            List<PropertyDescriptor> propertyDescriptors = getBeanPropertyDescriptor(beanClass);

            // Add all the bean properties as MethodProperties to this Item
            // later entries on the list overwrite earlier ones
            for (PropertyDescriptor pd : propertyDescriptors) {
                final Method getMethod = pd.getReadMethod();
                if ((getMethod != null)
                        && getMethod.getDeclaringClass() != Object.class) {
                    pdMap.put(pd.getName(), pd.getPropertyType());
                }
            }
        } catch (final java.beans.IntrospectionException ignored) {
        }

        return pdMap;
    }

    /**
     * Returns the property descriptors of a class or an interface.
     *
     * For an interface, superinterfaces are also iterated as Introspector does
     * not take them into account (Oracle Java bug 4275879), but in that case,
     * both the setter and the getter for a property must be in the same
     * interface and should not be overridden in subinterfaces for the discovery
     * to work correctly.
     *
     * For interfaces, the iteration is depth first and the properties of
     * superinterfaces are returned before those of their subinterfaces.
     *
     * @param beanClass
     * @return
     * @throws IntrospectionException
     */
    private static List<PropertyDescriptor> getBeanPropertyDescriptor(
            final Class<?> beanClass) throws IntrospectionException {
        // Oracle bug 4275879: Introspector does not consider superinterfaces of
        // an interface
        if (beanClass.isInterface()) {
            List<PropertyDescriptor> propertyDescriptors = new ArrayList<PropertyDescriptor>();

            for (Class<?> cls : beanClass.getInterfaces()) {
                propertyDescriptors.addAll(getBeanPropertyDescriptor(cls));
            }

            BeanInfo info = Introspector.getBeanInfo(beanClass);
            propertyDescriptors.addAll(Arrays.asList(info
                    .getPropertyDescriptors()));

            return propertyDescriptors;
        } else {
            BeanInfo info = Introspector.getBeanInfo(beanClass);
            return Arrays.asList(info.getPropertyDescriptors());
        }
    }
}
